"""Classes and methods for working with nondeterministic finite automata."""

from __future__ import annotations

from collections import deque
from itertools import chain, count, product, repeat
from typing import (
    AbstractSet,
    Any,
    Deque,
    Dict,
    FrozenSet,
    Generator,
    Iterable,
    List,
    Mapping,
    Optional,
    Set,
    Tuple,
    Type,
)

import networkx as nx
from cached_method import cached_method
from frozendict import frozendict
from typing_extensions import Self, TypeAlias

import automata.base.exceptions as exceptions
import automata.fa.dfa as dfa
import automata.fa.fa as fa
from automata.base.utils import get_reachable_nodes
from automata.regex.parser import RESERVED_CHARACTERS, parse_regex

NFAStateT = fa.FAStateT

NFAPathT = Mapping[str, AbstractSet[NFAStateT]]
NFATransitionsT = Mapping[NFAStateT, NFAPathT]
InputPathListT = List[Tuple[NFAStateT, NFAStateT, str]]


class NFA(fa.FA):
    """
    The `NFA` class is a subclass of `FA` and represents a nondeterministic finite
    automaton.

    Every NFA has the same five DFA properties: `state`, `input_symbols`,
    `transitions`, `initial_state`, and `final_states`. However, the structure of
    the `transitions` object has been modified slightly to accommodate the fact that
    a single state can have more than one transition for the same symbol. Therefore,
    instead of mapping a symbol to *one* end state in each sub-dict, each symbol is
    mapped to a *set* of end states.

    Every NFA can be rendered natively inside of a Jupyter notebook (automatically
    calling `show_diagram` without any arguments) if installed with the `visual`
    optional dependency.

    Parameters
    ----------
    states : AbstractSet[NFAStateT]
        Set of the NFA's valid states.
    input_symbols : AbstractSet[str]
        Set of the NFA's valid input symbols, each of which is a singleton
        string.
    transitions : Mapping[NFAStateT, Mapping[str, AbstractSet[NFAStateT]]]
        Dict consisting of the transitions for each state. Each key is a
        state name, and each value is another dict which maps a symbol
        (the key) to a set of states (the value).
    initial_state : NFAStateT
        The initial state for this NFA.
    final_states : AbstractSet[NFAStateT]
        A set of final states for this NFA.
    """

    __slots__: Tuple[str, ...] = (
        "states",
        "input_symbols",
        "transitions",
        "initial_state",
        "final_states",
        # These two entries are to allow for caching methods
        "__dict__",
        "__weakref__",
    )

    transitions: NFATransitionsT

    def __init__(
        self,
        *,
        states: AbstractSet[NFAStateT],
        input_symbols: AbstractSet[str],
        transitions: NFATransitionsT,
        initial_state: NFAStateT,
        final_states: AbstractSet[NFAStateT],
    ) -> None:
        """Initialize a complete NFA."""
        super().__init__(
            states=states,
            input_symbols=input_symbols,
            transitions=transitions,
            initial_state=initial_state,
            final_states=final_states,
        )

    @cached_method
    def _get_lambda_closures(self) -> Mapping[NFAStateT, FrozenSet[NFAStateT]]:
        """
        Computes a dictionary of the lambda closures for this NFA, where each
        key is the state name and the value is the lambda closure for that
        corresponding state. This dictionary is cached for the lifetime of the
        instance, and is available via the 'lambda_closures' attribute.

        The lambda closure of a state q is the set containing q, along with
        every state that can be reached from q by following only lambda
        transitions.
        """
        lambda_graph = nx.DiGraph()
        lambda_graph.add_nodes_from(self.states)
        lambda_graph.add_edges_from(
            (
                (start_state, end_state)
                for start_state, transition in self.transitions.items()
                for char, end_states in transition.items()
                if char == ""
                for end_state in end_states
            )
        )

        return frozendict(
            {
                state: frozenset(get_reachable_nodes(lambda_graph, [state]))
                for state in self.states
            }
        )

    def __add__(self, other: NFA) -> Self:
        """Return the concatenation of this NFA and another NFA."""
        if isinstance(other, NFA):
            return self.concatenate(other)
        else:
            return NotImplemented

    def __or__(self, other: NFA) -> Self:
        """Return the union of this NFA and another NFA."""
        if isinstance(other, NFA):
            return self.union(other)
        else:
            return NotImplemented

    def __and__(self, other: NFA) -> Self:
        """Return the union of this NFA and another NFA."""
        if isinstance(other, NFA):
            return self.intersection(other)
        else:
            return NotImplemented

    @classmethod
    def from_dfa(cls: Type[Self], target_dfa: dfa.DFA) -> Self:
        """
        Initialize this NFA as one equivalent to the given DFA.

        Parameters
        ----------
        target_dfa : DFA
            The DFA to construct an equivalent NFA for.

        Returns
        ------
        Self
            The NFA accepting the language of the input DFA.
        """
        nfa_transitions = {
            start_state: {
                input_symbol: {end_state} for input_symbol, end_state in paths.items()
            }
            for start_state, paths in target_dfa.transitions.items()
        }

        return cls(
            states=target_dfa.states,
            input_symbols=target_dfa.input_symbols,
            transitions=nfa_transitions,
            initial_state=target_dfa.initial_state,
            final_states=target_dfa.final_states,
        )

    def _validate_transition_invalid_symbols(
        self, start_state: NFAStateT, paths: NFAPathT
    ) -> None:
        """Raise an error if transition symbols are invalid."""
        for input_symbol in paths.keys():
            if input_symbol not in self.input_symbols and input_symbol != "":
                raise exceptions.InvalidSymbolError(
                    "state {} has invalid transition symbol {}".format(
                        start_state, input_symbol
                    )
                )

    @classmethod
    def from_regex(
        cls: Type[Self], regex: str, *, input_symbols: Optional[AbstractSet[str]] = None
    ) -> Self:
        """
        Initialize this NFA as one equivalent to the given regular expression.

        Parameters
        ----------
        regex : str
            The regex to construct an equivalent NFA for.
        input_symbols : Optional[AbstractSet[str]], default: None
            The set of input symbols to create the NFA over. If not
            set, defaults to all non-reserved characters found in the regex.

        Returns
        ------
        Self
            The NFA accepting the language of the input regex.
        """

        if input_symbols is None:
            input_symbols = frozenset(regex) - RESERVED_CHARACTERS
        else:
            conflicting_symbols = RESERVED_CHARACTERS & input_symbols
            if conflicting_symbols:
                raise exceptions.InvalidSymbolError(
                    f"Invalid input symbols: {conflicting_symbols}"
                )

        nfa_builder = parse_regex(regex, input_symbols)

        return cls(
            states=frozenset(nfa_builder._transitions.keys()),
            input_symbols=input_symbols,
            transitions=nfa_builder._transitions,
            initial_state=nfa_builder._initial_state,
            final_states=nfa_builder._final_states,
        )

    def _validate_transition_end_states(
        self, start_state: NFAStateT, paths: NFAPathT
    ) -> None:
        """Raise an error if transition end states are invalid."""
        for end_states in paths.values():
            for end_state in end_states:
                if end_state not in self.states:
                    raise exceptions.InvalidStateError(
                        "end state {} for transition on {} is not valid".format(
                            end_state, start_state
                        )
                    )

    def validate(self) -> None:
        """
        Raises an exception if this automaton is not internally consistent.

        Raises
        ------
        InvalidStateError
            If this NFA has invalid states in the transition dictionary.
        MissingStateError
            If this NFA has states missing from the transition dictionary.
        InvalidSymbolError
            If this NFA has invalid symbols in the transition dictionary.
        """
        for start_state, paths in self.transitions.items():
            self._validate_transition_invalid_symbols(start_state, paths)
            self._validate_transition_end_states(start_state, paths)
        self._validate_initial_state()
        self._validate_initial_state_transitions()
        self._validate_final_states()

    def _iterate_through_symbol_path_pairs(
        self, current_states: AbstractSet[NFAStateT]
    ) -> Generator[Tuple[str, FrozenSet[NFAStateT]], None, None]:
        """
        Iterate through input symbols with set of end transitions which are nonempty.
        """

        lambda_closures = self._get_lambda_closures()
        res_dict: Dict[str, Set[NFAStateT]] = {}

        for input_symbol, next_states in chain.from_iterable(
            self.transitions.get(state, {}).items() for state in current_states
        ):
            # Ignore empty string and empty end states
            if input_symbol and next_states:
                input_symbol_set = res_dict.setdefault(input_symbol, set())
                for end_state in next_states:
                    input_symbol_set.update(lambda_closures[end_state])

        for input_symbol, next_states in res_dict.items():
            yield (input_symbol, frozenset(next_states))

    def _get_next_current_states(
        self, current_states: AbstractSet[NFAStateT], input_symbol: str
    ) -> FrozenSet[NFAStateT]:
        """Return the next set of current states given the current set."""
        next_current_states: Set[NFAStateT] = set()
        lambda_closures = self._get_lambda_closures()

        for current_state in current_states:
            if current_state not in self.transitions:
                continue
            current_transition = self.transitions[current_state]
            for end_state in current_transition.get(input_symbol, {}):
                next_current_states.update(lambda_closures[end_state])

        return frozenset(next_current_states)

    @staticmethod
    def _compute_reachable_states(
        initial_state: NFAStateT, transitions: NFATransitionsT
    ) -> Set[NFAStateT]:
        """Compute the states which are reachable from the initial state."""

        visited_set: Set[NFAStateT] = set()
        queue: Deque[NFAStateT] = deque()

        queue.append(initial_state)
        visited_set.add(initial_state)

        while queue:
            state = queue.popleft()
            state_dict = transitions.get(state)

            if state_dict:
                for next_state in chain.from_iterable(
                    dest for dest in state_dict.values()
                ):
                    if next_state not in visited_set:
                        visited_set.add(next_state)
                        queue.append(next_state)

        return visited_set

    def _eliminate_lambda(
        self,
    ) -> Tuple[AbstractSet[NFAStateT], NFATransitionsT, AbstractSet[NFAStateT]]:
        """Internal helper function for eliminate lambda. Doesn't create final NFA."""

        # Create new transitions and final states for running this algorithm
        new_transitions = {
            state: {symbol: set(dest) for symbol, dest in paths.items()}
            for state, paths in self.transitions.items()
        }
        new_final_states = set(self.final_states)
        lambda_closures = self._get_lambda_closures()

        for state in self.states:
            lambda_enclosure = lambda_closures[state] - {state}
            for input_symbol in self.input_symbols:
                next_current_states = set(
                    self._get_next_current_states(lambda_enclosure, input_symbol)
                )

                # Don't do anything if no new current states
                if next_current_states:
                    state_transition_dict = new_transitions.setdefault(state, {})

                    if input_symbol in state_transition_dict:
                        state_transition_dict[input_symbol].update(next_current_states)
                    else:
                        state_transition_dict[input_symbol] = next_current_states

            if not new_final_states.isdisjoint(lambda_enclosure):
                new_final_states.add(state)

            if state in new_transitions:
                new_transitions[state].pop("", None)

        # Remove unreachable states
        reachable_states = self.__class__._compute_reachable_states(
            self.initial_state, new_transitions
        )
        reachable_final_states = reachable_states & new_final_states

        for state in self.states - reachable_states:
            new_transitions.pop(state, None)

        return reachable_states, new_transitions, reachable_final_states

    def eliminate_lambda(self) -> Self:
        """
        Returns an equivalent NFA with lambda transitions removed.

        Returns
        ------
        Self
            The equivalent NFA with lambda transitions removed.
        """

        (
            reachable_states,
            new_transitions,
            reachable_final_states,
        ) = self._eliminate_lambda()

        return self.__class__(
            states=reachable_states,
            input_symbols=self.input_symbols,
            transitions=new_transitions,
            initial_state=self.initial_state,
            final_states=reachable_final_states,
        )

    def _check_for_input_rejection(
        self, current_states: AbstractSet[NFAStateT]
    ) -> None:
        """Raise an error if the given config indicates rejected input."""
        if current_states.isdisjoint(self.final_states):
            raise exceptions.RejectionException(
                "the NFA stopped on all non-final states ({})".format(
                    ", ".join(str(state) for state in current_states)
                )
            )

    def read_input_stepwise(
        self, input_str: str
    ) -> Generator[AbstractSet[NFAStateT], None, None]:
        """
        Return a generator that yields the configuration of this NFA at each
        step while reading input.

        Parameters
        ----------
        input_str : str
            The input string to read.

        Yields
        ------
        Generator[AbstractSet[NFAStateT], None, None]
            A generator that yields the current configuration (a set of states) of
            the NFA after each step of reading input.

        Raises
        ------
        RejectionException
            Raised if this NFA does not accept the input string.
        """
        current_states = self._get_lambda_closures()[self.initial_state]

        yield current_states
        for input_symbol in input_str:
            current_states = self._get_next_current_states(current_states, input_symbol)
            yield current_states

        self._check_for_input_rejection(current_states)

    @staticmethod
    def _get_state_maps(
        state_set_a: AbstractSet[NFAStateT],
        state_set_b: AbstractSet[NFAStateT],
        *,
        start: int = 0,
    ) -> Tuple[Dict[NFAStateT, int], Dict[NFAStateT, int]]:
        """
        Generate state map dicts from given sets. Useful when the state set has
        to be a union of the state sets of component FAs.
        """

        state_name_counter = count(start)

        state_map_a = dict(zip(state_set_a, state_name_counter))
        state_map_b = dict(zip(state_set_b, state_name_counter))

        return (state_map_a, state_map_b)

    def union(self, other: NFA) -> Self:
        """
        Given two NFAs, M1 and M2, which accept the languages
        L1 and L2 respectively, returns an NFA which accepts
        the union of L1 and L2.

        Parameters
        ----------
        other : NFA
            The NFA we want to take a union with.

        Returns
        ------
        Self
            An NFA accepting the union of the two input NFAs.
        """

        # Starting at 1 because 0 is for the initial state
        (state_map_a, state_map_b) = self.__class__._get_state_maps(
            self.states, other.states, start=1
        )

        new_states = frozenset(chain(state_map_a.values(), state_map_b.values(), [0]))
        new_transitions: Dict[NFAStateT, Dict[str, Set[NFAStateT]]] = {
            state: {} for state in new_states
        }

        # Connect new initial state to both branch
        new_transitions[0] = {
            "": {state_map_a[self.initial_state], state_map_b[other.initial_state]}
        }

        # Transitions of self
        self.__class__._load_new_transition_dict(
            state_map_a, self.transitions, new_transitions
        )
        # Transitions of other
        self.__class__._load_new_transition_dict(
            state_map_b, other.transitions, new_transitions
        )

        # Final states
        new_final_states = frozenset(
            chain(
                (state_map_a[state] for state in self.final_states),
                (state_map_b[state] for state in other.final_states),
            )
        )

        return self.__class__(
            states=new_states,
            input_symbols=self.input_symbols | other.input_symbols,
            transitions=new_transitions,
            initial_state=0,
            final_states=new_final_states,
        )

    def concatenate(self, other: NFA) -> Self:
        """
        Given two NFAs, M1 and M2, which accept the languages
        L1 and L2 respectively, returns an NFA which accepts
        the language L1 concatenated with L2.

        Parameters
        ----------
        other : NFA
            The NFA we want to concatenate with.

        Returns
        ------
        Self
            An NFA accepting the language concatenation of the two input NFAs.
        """

        (state_map_a, state_map_b) = self.__class__._get_state_maps(
            self.states, other.states
        )

        new_states = frozenset(chain(state_map_a.values(), state_map_b.values()))
        new_transitions: Dict[NFAStateT, Dict[str, Set[NFAStateT]]] = {
            state: {} for state in new_states
        }

        # Transitions of self
        self.__class__._load_new_transition_dict(
            state_map_a, self.transitions, new_transitions
        )
        # Transitions of other
        self.__class__._load_new_transition_dict(
            state_map_b, other.transitions, new_transitions
        )

        # Transitions from self to other
        for state in self.final_states:
            new_transitions[state_map_a[state]].setdefault("", set()).add(
                state_map_b[other.initial_state]
            )

        # Final states of other
        new_final_states = frozenset(state_map_b[state] for state in other.final_states)

        return self.__class__(
            states=new_states,
            input_symbols=self.input_symbols | other.input_symbols,
            transitions=new_transitions,
            initial_state=state_map_a[self.initial_state],
            final_states=new_final_states,
        )

    def kleene_star(self) -> Self:
        """
        Given an NFA which accepts the language L, returns
        an NFA which accepts L repeated 0 or more times.

        Returns
        ------
        Self
            An NFA accepting the finite repetition of the input NFA.
        """
        new_states = set(self.states)
        new_initial_state = self.__class__._add_new_state(new_states)

        # Transitions are the same with a few additions.
        new_transitions: Dict[NFAStateT, Dict[str, Set[NFAStateT]]] = dict(
            self.transitions
        )
        # We add epsilon transition from new initial state
        # to old initial state
        new_transitions[new_initial_state] = {"": {self.initial_state}}
        # For each final state in original NFA we add epsilon
        # transition to the old initial state
        for state in self.final_states:
            new_transitions[state] = dict(new_transitions.get(state, {}))
            transition = new_transitions[state]
            transition[""] = set(transition.get("", set()))
            transition[""].add(self.initial_state)

        return self.__class__(
            states=new_states,
            input_symbols=self.input_symbols,
            transitions=new_transitions,
            initial_state=new_initial_state,
            final_states=self.final_states | {new_initial_state},
        )

    def option(self) -> Self:
        """
        Given an NFA which accepts the language L, returns
        an NFA which accepts L repeated 0 or 1 times.

        Returns
        ------
        Self
            An NFA accepting the optional of the input NFA.
        """
        new_states = set(self.states)
        new_initial_state = self.__class__._add_new_state(new_states)

        # Transitions are the same with a few additions.
        new_transitions = dict(self.transitions)
        # We add epsilon transition from new initial state
        # to old initial state
        new_transitions[new_initial_state] = {"": {self.initial_state}}

        return self.__class__(
            states=new_states,
            input_symbols=self.input_symbols,
            transitions=new_transitions,
            initial_state=new_initial_state,
            final_states=self.final_states | {new_initial_state},
        )

    def reverse(self) -> Self:
        """
        Given an NFA which accepts the language L,
        returns an NFA which accepts the reverse of L.

        Returns
        ------
        Self
            An NFA accepting the reversal of the input NFA.
        """
        new_states = set(self.states)
        new_initial_state = self.__class__._add_new_state(new_states)

        # Transitions are the same except reversed
        new_transitions: Dict[NFAStateT, Dict[str, Set[NFAStateT]]] = {
            state: {} for state in new_states
        }

        for state_a, transitions in self.transitions.items():
            for symbol, states in transitions.items():
                for state_b in states:
                    new_transitions[state_b].setdefault(symbol, set()).add(state_a)

        # And we additionally have epsilon transitions from
        # new initial state to each old final state.
        new_transitions[new_initial_state][""] = set(self.final_states)
        new_final_states = {self.initial_state}

        return self.__class__(
            states=new_states,
            input_symbols=self.input_symbols,
            transitions=new_transitions,
            initial_state=new_initial_state,
            final_states=new_final_states,
        )

    def intersection(self, other: NFA) -> Self:
        """
        Given two NFAs, M1 and M2, which accept the languages
        L1 and L2 respectively, returns an NFA which accepts
        the intersection of L1 and L2.

        Parameters
        ----------
        other : NFA
            The NFA we want to take an intersection with.

        Returns
        ------
        Self
            An NFA accepting the intersection of the two input NFAs.
        """

        new_states = set()
        new_input_symbols = self.input_symbols | other.input_symbols
        new_transitions: Dict[NFAStateT, Dict[str, Set[NFAStateT]]] = {}
        new_initial_state = (self.initial_state, other.initial_state)

        queue: Deque[NFAStateT] = deque()

        queue.append(new_initial_state)
        new_states.add(new_initial_state)

        while queue:
            curr_state = queue.popleft()
            q_a, q_b = curr_state

            # States we will consider adding to the queue
            next_states_iterables: List[Iterable[NFAStateT]] = list()

            # Get transition dict for states in self
            transitions_a = self.transitions.get(q_a, {})
            # Add epsilon transitions for first set of transitions
            epsilon_transitions_a = transitions_a.get("")
            if epsilon_transitions_a is not None:
                state_dict = new_transitions.setdefault(curr_state, {})
                state_dict.setdefault("", set()).update(
                    zip(epsilon_transitions_a, repeat(q_b))
                )
                next_states_iterables.append(zip(epsilon_transitions_a, repeat(q_b)))

            # Get transition dict for states in other
            transitions_b = other.transitions.get(q_b, {})
            # Add epsilon transitions for second set of transitions
            epsilon_transitions_b = transitions_b.get("")
            if epsilon_transitions_b is not None:
                state_dict = new_transitions.setdefault(curr_state, {})
                state_dict.setdefault("", set()).update(
                    zip(repeat(q_a), epsilon_transitions_b)
                )
                next_states_iterables.append(zip(repeat(q_a), epsilon_transitions_b))

            # Add all transitions moving over same input symbols
            for symbol in new_input_symbols:
                end_states_a = transitions_a.get(symbol)
                end_states_b = transitions_b.get(symbol)

                if end_states_a is not None and end_states_b is not None:
                    state_dict = new_transitions.setdefault(curr_state, {})
                    state_dict.setdefault(symbol, set()).update(
                        product(end_states_a, end_states_b)
                    )
                    next_states_iterables.append(product(end_states_a, end_states_b))

            # Finally, try visiting every state we found.
            for product_state in chain.from_iterable(next_states_iterables):
                if product_state not in new_states:
                    new_states.add(product_state)
                    queue.append(product_state)

        new_final_states = frozenset(
            (state_a, state_b)
            for (state_a, state_b) in new_states
            if state_a in self.final_states and state_b in other.final_states
        )

        return self.__class__(
            states=new_states,
            input_symbols=new_input_symbols,
            transitions=new_transitions,
            initial_state=new_initial_state,
            final_states=new_final_states,
        )

    def shuffle_product(self, other: NFA) -> Self:
        """
        Given two NFAs, M1 and M2, which accept the languages
        L1 and L2 respectively, returns an NFA which accepts
        the shuffle of L1 and L2.

        Parameters
        ----------
        other : NFA
            The NFA we want to take a shuffle product with.

        Returns
        ------
        Self
            An NFA accepting the shuffle product of the two input NFAs.
        """
        if not isinstance(other, NFA):
            raise TypeError(f"other must be an NFA, not {other.__class__.__name__}")

        new_input_symbols = self.input_symbols | other.input_symbols
        new_initial_state = (self.initial_state, other.initial_state)
        new_states = frozenset(product(self.states, other.states))

        new_transitions: Dict[NFAStateT, Dict[str, Set[NFAStateT]]] = {}

        for curr_state in new_states:
            state_dict = new_transitions.setdefault(curr_state, {})
            q_a, q_b = curr_state

            transitions_a = self.transitions.get(q_a, {})
            for symbol, end_states in transitions_a.items():
                state_dict.setdefault(symbol, set()).update(
                    zip(end_states, repeat(q_b))
                )

            transitions_b = other.transitions.get(q_b, {})
            for symbol, end_states in transitions_b.items():
                state_dict.setdefault(symbol, set()).update(
                    zip(repeat(q_a), end_states)
                )

        return self.__class__(
            states=new_states,
            input_symbols=new_input_symbols,
            transitions=new_transitions,
            initial_state=new_initial_state,
            final_states=frozenset(product(self.final_states, other.final_states)),
        )

    def right_quotient(self, other: NFA) -> Self:
        """
        Given two NFAs, M1 and M2, which accept the languages
        L1 and L2 respectively, returns an NFA which accepts
        the right quotient of L1 with respect to L2 (L1 / L2).

        Construction is based off of the one described here:
        https://cs.stackexchange.com/a/102043

        Parameters
        ----------
        other : NFA
            The NFA we want to take a right quotient with.

        Returns
        ------
        Self
            An NFA accepting the right quotient of the two input NFAs.
        """

        if not isinstance(other, NFA):
            raise TypeError(f"other must be an NFA, not {other.__class__.__name__}")

        # First, eliminate lambdas because they cause problems with this algorithm
        (
            self_reachable_states,
            self_new_transitions,
            self_reachable_final_states,
        ) = self._eliminate_lambda()
        (
            other_reachable_states,
            other_new_transitions,
            other_reachable_final_states,
        ) = other._eliminate_lambda()

        new_input_symbols = self.input_symbols | other.input_symbols
        new_initial_state = (self.initial_state, other.initial_state, False)
        new_final_states = frozenset(
            product(self_reachable_final_states, other_reachable_final_states, [True])
        )
        new_states = frozenset(
            chain(
                product(self_reachable_states, [other.initial_state], [False]),
                product(self_reachable_states, other_reachable_states, [True]),
            )
        )

        new_transitions: Dict[NFAStateT, Dict[str, Set[NFAStateT]]] = {}

        # Populate transitions for before reading the suffix
        for state in self_reachable_states:
            new_state = (state, other.initial_state, False)
            new_state_dict = new_transitions.setdefault(new_state, {})
            old_transitions_dict = self_new_transitions.get(state)

            if old_transitions_dict:
                for symbol, end_states in old_transitions_dict.items():
                    new_state_dict[symbol] = {
                        (end_state, other.initial_state, False)
                        for end_state in end_states
                    }

            new_state_dict[""] = {(state, other.initial_state, True)}

        # Start reading after the suffix
        for q_a, q_b in product(self_reachable_states, other_reachable_states):
            curr_state = (q_a, q_b, True)

            transitions_a = self_new_transitions.get(q_a, {})
            transitions_b = other_new_transitions.get(q_b, {})

            # Add all transitions moving over same input symbols
            for symbol in new_input_symbols:
                end_states_a = transitions_a.get(symbol)
                end_states_b = transitions_b.get(symbol)

                if end_states_a is not None and end_states_b is not None:
                    state_dict = new_transitions.setdefault(curr_state, {})
                    state_dict.setdefault("", set()).update(
                        product(end_states_a, end_states_b, [True])
                    )

        return self.__class__(
            states=new_states,
            input_symbols=new_input_symbols,
            transitions=new_transitions,
            initial_state=new_initial_state,
            final_states=new_final_states,
        )

    def left_quotient(self, other: NFA) -> Self:
        """
        Given two NFAs, M1 and M2, which accept the languages
        L1 and L2 respectively, returns an NFA which accepts
        the left quotient of L1 with respect to L2 (L2 \\ L1).

        Construction is based off of the one described here:
        https://cs.stackexchange.com/a/102043

        Parameters
        ----------
        other : NFA
            The NFA we want to take a left quotient with.

        Returns
        ------
        Self
            An NFA accepting the left quotient of the two input NFAs.
        """

        if not isinstance(other, NFA):
            raise TypeError(f"other must be an NFA, not {other.__class__.__name__}")

        # First, eliminate lambdas because they cause problems with this algorithm
        (
            self_reachable_states,
            self_new_transitions,
            self_reachable_final_states,
        ) = self._eliminate_lambda()
        (
            other_reachable_states,
            other_new_transitions,
            other_reachable_final_states,
        ) = other._eliminate_lambda()

        new_input_symbols = self.input_symbols | other.input_symbols
        new_initial_state = (self.initial_state, other.initial_state, False)
        new_final_states = frozenset(
            product(self_reachable_final_states, other_reachable_final_states, [True])
        )
        new_states = frozenset(
            chain(
                product(self_reachable_states, other_reachable_states, [False]),
                product(self_reachable_states, other_reachable_final_states, [True]),
            )
        )

        new_transitions: Dict[NFAStateT, Dict[str, Set[NFAStateT]]] = {}

        # Start reading the prefix
        for q_a, q_b in product(self_reachable_states, other_reachable_states):
            curr_state = (q_a, q_b, False)

            transitions_a = self_new_transitions.get(q_a, {})
            transitions_b = other_new_transitions.get(q_b, {})

            # Add all transitions moving over same input symbols
            for symbol in new_input_symbols:
                end_states_a = transitions_a.get(symbol)
                end_states_b = transitions_b.get(symbol)

                if end_states_a is not None and end_states_b is not None:
                    state_dict = new_transitions.setdefault(curr_state, {})
                    state_dict.setdefault("", set()).update(
                        product(end_states_a, end_states_b, [False])
                    )

            # Add lambda transition from final state, flipping third entry to true
            if q_b in other_reachable_final_states:
                state_dict = new_transitions.setdefault(curr_state, {})
                state_dict.setdefault("", set()).add((q_a, q_b, True))

        # Populate transitions for after reading the prefix
        for state_a, state_b in product(
            self_reachable_states, other_reachable_final_states
        ):
            new_state = (state_a, state_b, True)
            new_state_dict = new_transitions.setdefault(new_state, {})
            old_transitions_dict = self_new_transitions.get(state_a)

            if old_transitions_dict:
                for symbol, end_states in old_transitions_dict.items():
                    new_state_dict[symbol] = set(
                        zip(end_states, repeat(state_b), repeat(True))
                    )

        return self.__class__(
            states=new_states,
            input_symbols=new_input_symbols,
            transitions=new_transitions,
            initial_state=new_initial_state,
            final_states=new_final_states,
        )

    @staticmethod
    def _load_new_transition_dict(
        state_map_dict: Mapping[NFAStateT, NFAStateT],
        old_transition_dict: NFATransitionsT,
        new_transition_dict: Dict[NFAStateT, Dict[str, Set[NFAStateT]]],
    ) -> None:
        """
        Load the new_transition_dict with the old transitions corresponding to
        the given state_map_dict.
        """

        for state_a, transitions in old_transition_dict.items():
            for symbol, states in transitions.items():
                new_transition_dict[state_map_dict[state_a]][symbol] = {
                    state_map_dict[state_b] for state_b in states
                }

    def __eq__(self, other: Any) -> bool:
        """
        Return True if two NFAs are equivalent. Uses an optimized version of
        the extended Hopcroft-Karp algorithm (HKe). See
        https://arxiv.org/abs/0907.5058
        """

        NFAStatesPairT: TypeAlias = Tuple[FrozenSet[NFAStateT], int]

        # Must be another NFA and have equal alphabets
        if not isinstance(other, NFA) or self.input_symbols != other.input_symbols:
            return NotImplemented

        operand_nfas = (self, other)
        initial_state_a = (self._get_lambda_closures()[self.initial_state], 0)
        initial_state_b = (other._get_lambda_closures()[other.initial_state], 1)

        def is_final_state(states_pair: NFAStatesPairT) -> bool:
            states, operand_index = states_pair
            nfa = operand_nfas[operand_index]
            # If at least one of the current states is a final state, the
            # condition should satisfy
            return any(
                not nfa.final_states.isdisjoint(nfa._get_lambda_closures()[state])
                for state in states
            )

        def transition(states_pair: NFAStatesPairT, symbol: str) -> NFAStatesPairT:
            states, operand_index = states_pair
            return (
                operand_nfas[operand_index]._get_next_current_states(states, symbol),
                operand_index,
            )

        # Get data structures
        state_sets = nx.utils.union_find.UnionFind([initial_state_a, initial_state_b])
        pair_stack: Deque[Tuple[NFAStatesPairT, NFAStatesPairT]] = deque()

        # Do union find
        state_sets.union(initial_state_a, initial_state_b)
        pair_stack.append((initial_state_a, initial_state_b))

        while pair_stack:
            q_a, q_b = pair_stack.pop()

            if is_final_state(q_a) ^ is_final_state(q_b):
                return False

            for symbol in self.input_symbols:
                r_1 = state_sets[transition(q_a, symbol)]
                r_2 = state_sets[transition(q_b, symbol)]

                if r_1 != r_2:
                    state_sets.union(r_1, r_2)
                    pair_stack.append((r_1, r_2))

        return True

    @classmethod
    def edit_distance(
        cls: Type[Self],
        input_symbols: AbstractSet[str],
        reference_str: str,
        max_edit_distance: int,
        *,
        insertion: bool = True,
        deletion: bool = True,
        substitution: bool = True,
    ) -> Self:
        """
        Constructs the Levenshtein NFA for the given reference_str and given
        Levenshtein distance. This NFA recognizes strings within the given
        Levenshtein distance (commonly called edit distance) of the
        reference_str. Parameters control which error types the NFA will
        recognize (insertions, deletions, or substitutions). At least one error
        type must be set.

        If insertion and deletion are False and substitution is True, then this
        is the same as Hamming distance.

        If insertion and deletion are True and substitution is False, then this
        is the same as LCS distance.

        insertion, deletion, and substitution all default to True.

        Code adapted from:
        http://blog.notdot.net/2010/07/Damn-Cool-Algorithms-Levenshtein-Automata

        Parameters
        ----------
        input_symbols : AbstractSet[str]
            The set of input symbols to construct the NFA over.
        reference_str : str
            The reference string the NFA will use to recognize other close strings.
        max_edit_distance : int
            The maximum edit distance from the reference string this NFA will recognize.
            Must be positive.
        insertion : bool, default: True
            Whether to recognize insertion edits relative to the reference string.
        deletion : bool, default: True
            Whether to recognize deletion edits relative to the reference string.
        substitution : bool, default: True
            Whether to recognize substitution edits relative to the reference string.

        Returns
        ------
        Self
            An NFA accepting all strings within the given edit distance to the reference
            string.

        Raises
        ------
        ValueError
            Raised if the max_edit_distance is negative or all of the error flags are
            set to False (at least one must be True).
        """
        if max_edit_distance < 0:
            raise ValueError("max_edit_distance must be greater than zero")
        if not (insertion or deletion or substitution):
            raise ValueError(
                "At least one of insertion, deletion, or substitution must be enabled."
            )

        states = frozenset(
            product(range(len(reference_str) + 1), range(max_edit_distance + 1))
        )

        transitions: Dict[NFAStateT, Dict[str, Set[NFAStateT]]] = {}
        final_states = set()

        def add_transition(
            start_state_dict: Dict[str, Set[NFAStateT]],
            end_state: NFAStateT,
            symbol: str,
        ) -> None:
            """Add transition between start and end state on symbol"""
            start_state_dict.setdefault(symbol, set()).add(end_state)

        def add_any_transition(
            start_state_dict: Dict[str, Set[NFAStateT]], end_state: NFAStateT
        ) -> None:
            """Add transition on all symbols between start and end state"""
            for symbol in input_symbols:
                add_transition(start_state_dict, end_state, symbol)

        for (i, symbol), e in product(
            enumerate(reference_str), range(max_edit_distance + 1)
        ):
            state_transition_dict = transitions.setdefault((i, e), {})

            # Correct character
            add_transition(state_transition_dict, (i + 1, e), symbol)
            if e < max_edit_distance:
                if insertion:
                    # Insertion
                    add_any_transition(state_transition_dict, (i, e + 1))

                if deletion:
                    # Deletion
                    add_transition(state_transition_dict, (i + 1, e + 1), "")

                if substitution:
                    # Substitution
                    add_any_transition(state_transition_dict, (i + 1, e + 1))

        for e in range(max_edit_distance + 1):
            state_transition_dict = transitions.setdefault((len(reference_str), e), {})
            if insertion and e < max_edit_distance:
                add_any_transition(state_transition_dict, (len(reference_str), e + 1))

            final_states.add((len(reference_str), e))

        return cls(
            states=states,
            input_symbols=input_symbols,
            transitions=transitions,
            initial_state=(0, 0),
            final_states=final_states,
        )

    def iter_transitions(
        self,
    ) -> Generator[Tuple[NFAStateT, NFAStateT, str], None, None]:
        """
        Iterate over all transitions in the NFA. Each transition is a tuple
        of the form (from_state, to_state, symbol).

        Returns
        ------
        Generator[Tuple[NFAStateT, NFAStateT, str], None, None]
            The desired generator over the NFA transitions.
        """
        return (
            (from_, to_, symbol)
            for from_, lookup in self.transitions.items()
            for symbol, to_lookup in lookup.items()
            for to_ in to_lookup
        )

    def _get_input_path(self, input_str: str) -> Tuple[InputPathListT, bool]:
        """
        Get best input path. A path is better if (with priority):

        1. It is an accepting path (ends in a final state)
        2. It reads more of the input (if the input is not accepted we
            select the path such that we can stay on the nfa the longest)
        3. It has the fewest jumps (uses less lambda symbols)

        Returns a tuple of:
        1. the path taken
        2. whether the path was accepting
        """

        visited = set()
        work_queue: Deque[Tuple[InputPathListT, NFAStateT, str]] = deque(
            [([], self.initial_state, input_str)]
        )

        last_non_accepting_input: InputPathListT = []
        least_input_remaining = len(input_str)

        while work_queue:
            visited_states, curr_state, remaining_input = work_queue.popleft()

            # First final state we hit is the best according to desired criteria
            if curr_state in self.final_states and not remaining_input:
                return visited_states, True

            # Otherwise, update longest non-accepting input
            if len(remaining_input) < least_input_remaining:
                least_input_remaining = len(remaining_input)
                last_non_accepting_input = visited_states

            # First, get next states that result from reading from input
            if remaining_input:
                next_symbol = remaining_input[0]
                rest = remaining_input[1:] if remaining_input else ""

                next_states_from_symbol = self.transitions[curr_state].get(
                    next_symbol, set()
                )

                for next_state in next_states_from_symbol:
                    if (next_state, rest) not in visited:
                        next_visited_states = visited_states.copy()
                        next_visited_states.append(
                            (curr_state, next_state, next_symbol)
                        )
                        visited.add((next_state, rest))
                        work_queue.append((next_visited_states, next_state, rest))

            # Next, get next states resulting from lambda transition
            next_states_from_lambda = self.transitions[curr_state].get("", set())

            for next_state in next_states_from_lambda:
                if (next_state, remaining_input) not in visited:
                    next_visited_states = visited_states.copy()
                    next_visited_states.append((curr_state, next_state, ""))
                    visited.add((next_state, remaining_input))
                    work_queue.append(
                        (next_visited_states, next_state, remaining_input)
                    )

        return last_non_accepting_input, False



"""Classes and methods for working with Turing machine configurations."""

from dataclasses import dataclass
from typing import Tuple

from automata.tm.tape import TMTape
from automata.tm.tm import TMStateT


@dataclass(frozen=True)
class TMConfiguration:
    """A Turing machine configuration."""

    __slots__ = ("state", "tape")

    state: TMStateT
    tape: TMTape

    def __repr__(self) -> str:
        """Return a string representation of the configuration."""
        return "{}({!r}, {!r})".format(self.__class__.__name__, self.state, self.tape)

    def print(self) -> None:
        """Print the machine's current configuration in a readable form."""
        print(
            "{current_state}: {tape}\n{current_position}".format(
                current_state=self.state,
                tape="".join(self.tape).rjust(len(self.tape), self.tape.blank_symbol),
                current_position="^".rjust(
                    self.tape.current_position + len(self.state) + 3
                ),
            )
        )


@dataclass(frozen=True)
class MTMConfiguration:
    """A Multitape Turing machine configuration."""

    __slots__ = ("state", "tapes")

    state: TMStateT
    tapes: Tuple[TMTape, ...]

    def __repr__(self) -> str:
        """Return a string representation of the configuration."""
        return "{}({!r}, {!r})".format(self.__class__.__name__, self.state, self.tapes)

    def print(self) -> None:
        """Print the machine's current configuration in a readable form."""
        description = "{}: \n".format(self.state)
        for i, tape in enumerate(self.tapes, start=1):
            title = "> Tape {}: ".format(i)
            position = tape.current_position + len(title) + 1
            description += "> Tape {j}: {tape}\n{current_position}\n".format(
                j=i,
                tape="".join(tape).ljust(tape.current_position, "#"),
                current_position="^".rjust(position),
            )
        print(description)



"""Classes and methods for working with all Turing machines."""

import abc
from typing import AbstractSet, Literal

import automata.base.exceptions as exceptions
from automata.base.automaton import Automaton, AutomatonStateT

TMStateT = AutomatonStateT
TMDirectionT = Literal["L", "R", "N"]


class TM(Automaton, metaclass=abc.ABCMeta):
    """An abstract base class for Turing machines."""

    __slots__ = tuple()

    tape_symbols: AbstractSet[str]
    blank_symbol: str

    def _read_input_symbol_subset(self) -> None:
        if not (self.input_symbols < self.tape_symbols):
            raise exceptions.MissingSymbolError(
                "The set of tape symbols is missing symbols from the input "
                "symbol set ({})".format(self.tape_symbols - self.input_symbols)
            )

    def _validate_blank_symbol(self) -> None:
        """Raise an error if blank symbol is not a tape symbol."""
        if self.blank_symbol not in self.tape_symbols:
            raise exceptions.InvalidSymbolError(
                "blank symbol {} is not a tape symbol".format(self.blank_symbol)
            )

    def _validate_nonfinal_initial_state(self) -> None:
        """Raise an error if the initial state is a final state."""
        if self.initial_state in self.final_states:
            raise exceptions.InitialStateError(
                "initial state {} cannot be a final state".format(self.initial_state)
            )



"""Functions for testing the global Automata configuration."""

import unittest
from unittest.mock import MagicMock, patch

from frozendict import frozendict

import automata.base.config as global_config
from automata.fa.dfa import DFA


class TestConfig(unittest.TestCase):
    orig_should_validate: bool
    orig_allow_mutable_automata: bool

    def setUp(self) -> None:
        self.orig_should_validate = global_config.should_validate_automata
        self.orig_allow_mutable_automata = global_config.allow_mutable_automata

    def tearDown(self) -> None:
        global_config.should_validate_automata = self.orig_should_validate
        global_config.allow_mutable_automata = self.orig_allow_mutable_automata

    @patch("automata.fa.dfa.DFA.validate")
    def test_disable_validation(self, validate: MagicMock) -> None:
        """Should disable automaton validation"""
        global_config.should_validate_automata = False
        DFA.universal_language({"0", "1"})
        validate.assert_not_called()

    @patch("automata.base.utils.freeze_value")
    def test_disable_ensure_values_are_frozen(self, freeze_value: MagicMock) -> None:
        """Should enable automaton mutability"""
        global_config.allow_mutable_automata = True
        DFA(
            states=frozenset(["s1"]),
            input_symbols=frozenset("a"),
            transitions=frozendict({"s1": frozendict({"a": "s1"})}),
            initial_state="s1",
            final_states=frozenset(["s1"]),
        )
        freeze_value.assert_not_called()

        # Also this should not call freeze_value nor throw any error
        DFA(
            states={"s1"},
            input_symbols={"a"},
            transitions={"s1": {"a": "s1"}},
            initial_state="s1",
            final_states={"s1"},
        )
        freeze_value.assert_not_called()



"""Classes and methods for working with PDA configurations."""

from dataclasses import dataclass
from typing import Any

from automata.base.automaton import AutomatonStateT
from automata.pda.stack import PDAStack


@dataclass(frozen=True)
class PDAConfiguration:
    """
    A configuration is a triple of current state, remaining input and stack.

    It represents the complete runtime state of a PDA.
    It is hashable and immutable.
    """

    __slots__ = ("state", "remaining_input", "stack")

    state: AutomatonStateT
    remaining_input: str
    stack: PDAStack

    def __repr__(self) -> str:
        """Return a string representation of the configuration."""
        return "{}({!r}, {!r}, {!r})".format(
            self.__class__.__name__, self.state, self.remaining_input, self.stack
        )

    def __eq__(self, other: Any) -> bool:
        """Return True if two PDAConfiguration are equivalent"""
        if not isinstance(other, PDAConfiguration):
            return NotImplemented

        return (
            self.state == other.state
            and self.remaining_input == other.remaining_input
            and self.stack == other.stack
        )



"""Classes and methods for working with Turing machine tapes."""

from dataclasses import dataclass
from typing import Iterator, Sequence, Tuple

from typing_extensions import Self

from automata.tm.tm import TMDirectionT


@dataclass(frozen=True)
class TMTape:
    """A Turing machine tape."""

    __slots__ = ("tape", "blank_symbol", "current_position")

    tape: Tuple[str]
    blank_symbol: str
    current_position: int

    def __init__(
        self, tape: Sequence[str], blank_symbol: str, current_position: int = 0
    ) -> None:
        """Initialize a new Turing machine tape."""
        tape = list(tape)
        # Make sure that there's something under the cursor.
        while len(tape) <= current_position:
            tape.append(blank_symbol)
        object.__setattr__(self, "tape", tuple(tape))
        object.__setattr__(self, "blank_symbol", blank_symbol)
        object.__setattr__(self, "current_position", current_position)

    def read_symbol(self) -> str:
        """
        Get the symbol at the current position in the tape.

        Returns
        ------
        str
            The symbol at the current position in the tape.
        """

        return self.tape[self.current_position]

    def write_symbol(self, new_tape_symbol: str) -> Self:
        """
        Write the given symbol at the current position in the tape.

        Parameters
        ------
        new_tape_symbol: str
            The new symbol to write onto the tape.

        Returns
        ------
        TMTape
            A copy of the TMTape with the new symbol written at the current position.
        """

        tape_elements = list(self.tape)
        tape_elements[self.current_position] = new_tape_symbol
        return self.__class__(
            tape_elements,
            blank_symbol=self.blank_symbol,
            current_position=self.current_position,
        )

    def move(self, direction: TMDirectionT) -> Self:
        """
        Move the tape to the next symbol in the given direction.

        Parameters
        ------
        direction: TMDirectionT
            The direction to move the current position of the tape.

        Returns
        ------
        TMTape
            A copy of the TMTape with the current position moved.
        """

        # Copy stuff.
        new_tape = list(self.tape)
        new_position = self.current_position
        if direction == "R":
            new_position += 1
        elif direction == "N":
            pass
        elif direction == "L":
            new_position -= 1
        # Make sure that the cursor doesn't run off the end of the tape.
        if new_position == -1:
            new_tape.insert(0, self.blank_symbol)
            new_position += 1
        if new_position == len(new_tape):
            new_tape.append(self.blank_symbol)
        return self.__class__(
            new_tape, blank_symbol=self.blank_symbol, current_position=new_position
        )

    def copy(self) -> Self:
        """
        Make a copy of this TMTape.

        Returns
        ------
        TMTape
            A copy of the TMTape.
        """
        return self.__class__(
            list(self.tape).copy(),
            blank_symbol=self.blank_symbol,
            current_position=self.current_position,
        )

    def get_symbols_as_str(self) -> str:
        return "".join(self.tape)

    def __len__(self) -> int:
        """Return the number of symbols on the tape."""
        return len(self.tape)  # TODO: do we count the blank symbols?

    def __iter__(self) -> Iterator[str]:
        """Return an iterator for the tape."""
        return iter(self.tape)

    def __repr__(self) -> str:
        """Return a string representation of the tape."""
        return "{}({!r}, {!r}, {!r})".format(
            self.__class__.__name__,
            "".join(self.tape),
            self.blank_symbol,
            self.current_position,
        )



"""Exception classes specific to Turing machines."""

from automata.base.exceptions import AutomatonException


class TMException(AutomatonException):
    """The base class for all machine-related errors."""

    pass


class InvalidDirectionError(TMException):
    """A direction is not a valid direction for this machine."""

    pass


class InconsistentTapesException(TMException):
    """The number of tapes defined for the multitape Turing machine is
    not consistent with the definitions of the transitions."""

    pass


class MalformedExtendedTapeError(TMException):
    """Extended tape for simulating an mntm as a ntm is not valid.
    Either there are 2 virtual heads for a virtual tape or the
    a head symbol is at the leftmost end of a virtual tape."""

    pass



"""Classes and functions for testing the behavior of PDAs."""

import tempfile
import unittest

from automata.pda.dpda import DPDA
from automata.pda.npda import NPDA
from automata.pda.pda import PDA


class TestPDA(unittest.TestCase):
    """A test class for testing all pushdown automata."""

    dpda: DPDA
    npda: NPDA

    temp_dir_path = tempfile.gettempdir()

    def setUp(self) -> None:
        """Reset test automata before every test function."""
        # DPDA which which matches zero or more 'a's, followed by the same
        # number of 'b's (accepting by final state)
        self.dpda = DPDA(
            states={"q0", "q1", "q2", "q3"},
            input_symbols={"a", "b"},
            stack_symbols={"0", "1"},
            transitions={
                "q0": {"a": {"0": ("q1", ("1", "0"))}},
                "q1": {"a": {"1": ("q1", ("1", "1"))}, "b": {"1": ("q2", "")}},
                "q2": {"b": {"1": ("q2", "")}, "": {"0": ("q3", ("0",))}},
            },
            initial_state="q0",
            initial_stack_symbol="0",
            final_states={"q3"},
            acceptance_mode="final_state",
        )

        # NPDA which matches palindromes consisting of 'a's and 'b's
        # (accepting by final state)
        # q0 reads the first half of the word, q1 the other half, q2 accepts.
        # But we have to guess when to switch.
        self.npda = NPDA(
            states={"q0", "q1", "q2"},
            input_symbols={"a", "b"},
            stack_symbols={"A", "B", "#"},
            transitions={
                "q0": {
                    "": {
                        "#": {("q2", "#")},
                    },
                    "a": {
                        "#": {("q0", ("A", "#"))},
                        "A": {
                            ("q0", ("A", "A")),
                            ("q1", ""),
                        },
                        "B": {("q0", ("A", "B"))},
                    },
                    "b": {
                        "#": {("q0", ("B", "#"))},
                        "A": {("q0", ("B", "A"))},
                        "B": {
                            ("q0", ("B", "B")),
                            ("q1", ""),
                        },
                    },
                },
                "q1": {
                    "": {"#": {("q2", "#")}},
                    "a": {"A": {("q1", "")}},
                    "b": {"B": {("q1", "")}},
                },
            },
            initial_state="q0",
            initial_stack_symbol="#",
            final_states={"q2"},
            acceptance_mode="final_state",
        )


class TestPDAAbstract(unittest.TestCase):
    def test_abstract_methods_not_implemented(self) -> None:
        """Should raise NotImplementedError when calling abstract methods."""
        abstract_methods = {
            "iter_transitions": (PDA,),
            "_get_input_path": (PDA, ""),
        }
        for method_name, method_args in abstract_methods.items():
            with self.assertRaises(NotImplementedError):
                getattr(PDA, method_name)(*method_args)



"""Classes and methods for working with multitape nondeterministic Turing
machines."""

from collections import deque
from itertools import chain
from typing import AbstractSet, Any, Generator, Mapping, Optional, Sequence, Tuple

import automata.base.exceptions as exceptions
import automata.tm.exceptions as tm_exceptions
import automata.tm.ntm as ntm
import automata.tm.tm as tm
from automata.tm.configuration import MTMConfiguration, TMConfiguration
from automata.tm.tape import TMTape

MNTMStateT = ntm.NTMStateT
MNTMPathResultT = Tuple[str, Tuple[Tuple[str, tm.TMDirectionT], ...]]
MNTMPathT = Mapping[Tuple[str, ...], Sequence[MNTMPathResultT]]
MNTMTransitionsT = Mapping[str, MNTMPathT]


class MNTM(ntm.NTM):
    """
    The `MNTM` class is a subclass of `TM` and represents a multitape
    (non)deterministic Turing machine.

    Parameters
    ----------
    states: AbstractSet[MNTMStateT]
        A set of the MNTM's valid states.
    input_symbols: AbstractSet[str]
        Set of the MNTM's valid input symbols, each of which is a singleton
        string.
    tape_symbols: AbstractSet[str]
        Set of the MNTM's valid tape symbols, each of which is a singleton
        string.
    n_tapes: int
        The number of tapes in this MNTM.
    transitions: MNTMTransitionsT
        Dict consisting of the transitions for each state; each key is a
        state name, and each value is a dict which maps a symbol (the key) to
        a list of tuples consisting of the next state, the symbol to write on the
        tape, and the direction to move the tape head.
    initial_state: MNTMStateT
        The name of the initial state for this MNTM.
    blank_symbol: str
        A symbol from `tape_symbols` to be used as the blank symbol
        for this MNTM.
    final_states: AbstractSet[MNTMStateT]
        A set of final states for this MNTM.

    Example
    ----------
        from automata.tm.mntm import MNTM
        # MNTM which accepts all strings in {0, 1}* and writes all
        # 1's from the first tape (input) to the second tape.
        self.mntm1 = MNTM(
            states={'q0', 'q1'},
            input_symbols={'0', '1'},
            tape_symbols={'0', '1', '#'},
            n_tapes=2,
            transitions={
                'q0': {
                    ('1', '#'): [('q0', (('1', 'R'), ('1', 'R')))],
                    ('0', '#'): [('q0', (('0', 'R'), ('#', 'N')))],
                    ('#', '#'): [('q1', (('#', 'N'), ('#', 'N')))],
                }
            },
            initial_state='q0',
            blank_symbol='#',
            final_states={'q1'},
        )
    """

    __slots__ = (
        "states",
        "input_symbols",
        "tape_symbols",
        "n_tapes",
        "transitions",
        "initial_state",
        "blank_symbol",
        "final_states",
    )

    transitions: MNTMTransitionsT  # type: ignore
    n_tapes: int

    def __init__(
        self,
        *,
        states: AbstractSet[MNTMStateT],
        input_symbols: AbstractSet[str],
        tape_symbols: AbstractSet[str],
        n_tapes: int,
        transitions: MNTMTransitionsT,
        initial_state: MNTMStateT,
        blank_symbol: str,
        final_states: AbstractSet[MNTMStateT],
    ):
        """Initialize a complete Turing machine."""
        super(tm.TM, self).__init__(
            states=states,
            input_symbols=input_symbols,
            tape_symbols=tape_symbols,
            transitions=transitions,
            initial_state=initial_state,
            blank_symbol=blank_symbol,
            final_states=final_states,
            n_tapes=n_tapes,
        )

    def _validate_transition_symbols(self, state: MNTMStateT, paths: Mapping) -> None:
        for tape_symbol in [
            tape_symbol for symbol in paths.keys() for tape_symbol in symbol
        ]:
            if tape_symbol not in self.tape_symbols:
                raise exceptions.InvalidSymbolError(
                    "transition symbol {} for state {} is not valid".format(
                        tape_symbol, state
                    )
                )

    def _validate_transition_state(self, transition_state: MNTMStateT) -> None:
        if transition_state not in self.states:
            raise exceptions.InvalidStateError(
                "transition state is not valid ({})".format(transition_state)
            )

    def _validate_transition_results(self, paths: Mapping) -> None:
        for results in paths.values():
            for result in results:
                state, moves = result
                for move in moves:
                    symbol, direction = move
                    possible_result = (state, symbol, direction)
                    self._validate_transition_result(possible_result)

    def _validate_tapes_consistency(self) -> None:
        for state in self.transitions:
            for read_tape_symbols in self.transitions[state]:
                if len(read_tape_symbols) != self.n_tapes:
                    error = (
                        "tapes symbols {} inconsistent with the number of "
                        "tapes defined. Expected {} symbols, got {}"
                    ).format(read_tape_symbols, self.n_tapes, len(read_tape_symbols))
                    raise tm_exceptions.InconsistentTapesException(error)
                for transition in self.transitions[state][read_tape_symbols]:
                    _, moves = transition
                    if len(moves) != self.n_tapes:
                        error = (
                            "transition {} has inconsistent operations on "
                            "tapes. Expected {} write/move operations, "
                            "got {}"
                        ).format(transition, self.n_tapes, len(moves))
                        raise tm_exceptions.InconsistentTapesException(error)

    def validate(self) -> None:
        """
        Raises an exception if this automaton is not internally consistent.

        Raises
        ------
        InvalidStateError
            If this MNTM has invalid states in the transition dictionary.
        InvalidSymbolError
            If this MNTM has invalid symbols in the transition dictionary.
        InvalidDirectionError
            If this MNTM has a transition with an invalid direction.
        FinalStateError
            If this MNTM has a transition on any final states.
        InconsistentTapesException
            If this MNTM has inconsistent tape contents.
        """
        super().validate()
        self._validate_tapes_consistency()

    def _get_tapes_for_input_str(self, input_str: str) -> Tuple[TMTape, ...]:
        """Produce a new list of tapes based on an input string to be read."""
        return (
            # Input is saved on first tape
            TMTape(input_str, blank_symbol=self.blank_symbol),
            # The rest of the tapes have blanks
            *(
                TMTape(
                    self.blank_symbol,
                    blank_symbol=self.blank_symbol,
                    current_position=0,
                )
                for _ in range(self.n_tapes - 1)
            ),
        )

    def _read_current_tape_symbols(self, tapes: Tuple[TMTape, ...]) -> Tuple[str, ...]:
        """Reads the current tape symbols in each of the tapes and their
        corresponding heads."""
        return tuple(tape.read_symbol() for tape in tapes)

    def _get_transition(
        self, current_state: MNTMStateT, tapes: Tuple[TMTape, ...]
    ) -> Optional[Sequence[MNTMPathResultT]]:
        """Get the transition tuple for the given state and tape symbols in
        each tape."""
        current_tape_symbols = self._read_current_tape_symbols(tapes)
        if (
            current_state in self.transitions
            and current_tape_symbols in self.transitions[current_state]
        ):
            return self.transitions[current_state][
                self._read_current_tape_symbols(tapes)
            ]
        else:
            return None

    def _get_next_configuration(
        self, transition: MNTMPathResultT, current_tapes: Tuple[TMTape, ...]
    ) -> MTMConfiguration:
        """Advances to the next configuration."""
        current_state, moves = transition
        tapes = tuple(
            tape.write_symbol(symbol).move(direction)
            for (symbol, direction), tape in zip(moves, current_tapes)
        )

        return MTMConfiguration(state=current_state, tapes=tapes)

    def read_input_stepwise(self, input_str: str) -> Generator[Any, None, Any]:
        """
        Checks if the given string is accepted by this MNTM machine,
        using a BFS of every possible configuration from each configuration.
        Yields the current configuration of the machine at each step.

        Parameters
        ----------
        input_str : str
            The input string to read.

        Yields
        ------
        Generator[Set[MTMConfiguration], None, None]
            A generator that yields the current configuration of
            the DTM after each step of reading input.

        Raises
        ------
        RejectionException
            Raised if this MNTM does not accept the input string.
        """
        # TODO Any type above should be Set[MTMConfiguration], refactor required

        tapes = self._get_tapes_for_input_str(input_str)
        queue = deque([(MTMConfiguration(state=self.initial_state, tapes=tapes))])
        while len(queue) > 0:
            current_config = queue.popleft()
            yield {MTMConfiguration(current_config.state, current_config.tapes)}

            possible_transitions = self._get_transition(
                current_config.state, current_config.tapes
            )
            if possible_transitions is None:
                if current_config.state in self.final_states:
                    return {
                        MTMConfiguration(current_config.state, current_config.tapes)
                    }
            else:
                for transition in possible_transitions[1:]:
                    queue.append(
                        self._get_next_configuration(
                            transition, current_tapes=current_config.tapes
                        )
                    )

                queue.append(
                    self._get_next_configuration(
                        possible_transitions[0], current_tapes=current_config.tapes
                    )
                )

        raise exceptions.RejectionException(
            "the multitape MNTM did not reach an accepting configuration"
        )

    @staticmethod
    def _read_extended_tape(
        tape: str, head_symbol: str = "^", tape_separator_symbol: str = "_"
    ) -> Tuple[str, ...]:
        """
        Returns a tuple with the symbols extracted from the given
        tape, that are the virtual heads for their corresponding
        virtual tape.
        """

        virtual_heads = []
        heads_found = 0
        separators_found = 0
        for i, symbol in enumerate(tape):
            if symbol == head_symbol:
                if i - 1 < 0:
                    raise tm_exceptions.MalformedExtendedTapeError(
                        "head symbol was found on leftmost end of the extended tape"
                    )
                else:
                    previous_symbol = tape[i - 1]
                    virtual_heads.append(previous_symbol)
                    heads_found += 1
            elif symbol == tape_separator_symbol:
                if heads_found == 0:
                    raise tm_exceptions.MalformedExtendedTapeError(
                        "no head symbol found on one of the virtual tapes"
                    )
                elif heads_found > 1:
                    raise tm_exceptions.MalformedExtendedTapeError(
                        "more than one head symbol found on one of the virtual tapes"
                    )
                else:
                    heads_found = 0
                    separators_found += 1

        if len(virtual_heads) != separators_found:
            raise tm_exceptions.MalformedExtendedTapeError(
                "there must be 1 virtual head for every tape separator symbol"
            )

        return tuple(virtual_heads)

    def read_input_as_ntm(
        self, input_str: str
    ) -> Generator[AbstractSet[TMConfiguration], None, None]:
        """
        Simulates this MNTM as a single-tape Turing machine.
        Yields the configuration at each step.

        Parameters
        ----------
        input_str : str
            The input string to read.

        Yields
        ------
        Generator[AbstractSet[TMConfiguration], None, None]
            A generator that yields the current configuration of
            the MNTM as a set after each step of reading input.

        Raises
        ------
        RejectionException
            Raised if this MNTM does not accept the input string.

        """
        tapes = self._get_tapes_for_input_str(input_str)
        head_symbol = "^"
        tape_separator_symbol = "_"

        # Make string from all tapes
        initial_tape = "".join(
            chain.from_iterable(
                (tape.tape[0], head_symbol, *tape.tape[1:], tape_separator_symbol)
                for tape in tapes
            )
        )

        # Use a queue for BFS to handle non-determinism
        # (state, tape, position)
        queue = deque([(self.initial_state, initial_tape, 0)])

        while queue:
            current_state, extended_tape, pos = queue.popleft()
            current_config = TMConfiguration(
                current_state,
                TMTape(
                    extended_tape, blank_symbol=self.blank_symbol, current_position=pos
                ),
            )

            yield {current_config}

            if current_state in self.final_states:
                return

            # Read virtual heads from the tape
            virtual_heads = self._read_extended_tape(
                extended_tape, head_symbol, tape_separator_symbol
            )

            # Get all possible transitions
            try:
                possible_configs = self.transitions[current_state][virtual_heads]
            except KeyError:
                continue  # No valid transition, try other paths

            # Process each possible next configuration
            for next_config in possible_configs:
                next_state, moves = next_config
                new_tape = extended_tape
                i = 0  # current position

                # Apply the moves to get the next tape configuration
                for move in moves:
                    new_head, direction = move
                    executing_changes = True

                    while executing_changes:
                        if new_tape[i] == head_symbol:
                            # Update the tape symbol before the head
                            new_tape = new_tape[: i - 1] + new_head + new_tape[i:]
                            # Remove the old head
                            new_tape = new_tape[:i] + "" + new_tape[i + 1 :]

                            # Move the head according to direction
                            if direction == "R":
                                i += 1
                            elif direction == "L":
                                i -= 1
                            # else direction == 'N', i stays the same

                            # Handle edge cases with tape separator
                            if i > 0 and new_tape[i - 1] == tape_separator_symbol:
                                i -= 1
                                new_tape = (
                                    new_tape[:i]
                                    + self.blank_symbol
                                    + head_symbol
                                    + new_tape[i:]
                                )
                                i += 1
                            else:
                                new_tape = new_tape[:i] + head_symbol + new_tape[i:]

                        elif new_tape[i] == tape_separator_symbol:
                            executing_changes = False

                        i += 1

                # Add the new configuration to the queue
                queue.append((next_state, new_tape, i - 1))

        raise exceptions.RejectionException(
            "the multitape NTM did not reach an accepting configuration"
        )



"""Functions for testing the Automaton abstract base class."""

import unittest

from automata.base.automaton import Automaton


class TestAutomaton(unittest.TestCase):
    def test_abstract_methods_not_implemented(self) -> None:
        """Should raise NotImplementedError when calling abstract methods."""
        abstract_methods = {
            "validate": (Automaton,),
            "read_input_stepwise": (Automaton, ""),
        }
        for method_name, method_args in abstract_methods.items():
            with self.assertRaises(NotImplementedError):
                getattr(Automaton, method_name)(*method_args)



"""Classes for working with all automata, including Turing machines."""

import abc
from typing import AbstractSet, Any, Dict, Generator, Mapping, NoReturn, Tuple

from frozendict import frozendict
from typing_extensions import Self

import automata.base.config as global_config
import automata.base.exceptions as exceptions
from automata.base.utils import freeze_value

AutomatonStateT = Any
AutomatonPathT = Mapping[str, Any]
AutomatonTransitionsT = Mapping[str, AutomatonPathT]


class Automaton(metaclass=abc.ABCMeta):
    """An abstract base class for all automata, including Turing machines."""

    __slots__: Tuple[str, ...] = tuple()

    initial_state: AutomatonStateT
    states: AbstractSet[AutomatonStateT]
    final_states: AbstractSet[AutomatonStateT]
    transitions: AutomatonTransitionsT
    input_symbols: AbstractSet[str]

    @staticmethod
    def _get_state_name(state_data: Any) -> str:
        """
        Get a string representation of a state. This is used for displaying and
        uses `str` for any unsupported python data types.
        """
        if isinstance(state_data, str):
            if state_data == "":
                return "λ"

            return state_data

        elif isinstance(state_data, (frozenset, tuple)):
            inner = ", ".join(
                Automaton._get_state_name(sub_data) for sub_data in state_data
            )
            if isinstance(state_data, frozenset):
                if state_data:
                    return "{" + inner + "}"
                else:
                    return "∅"

            elif isinstance(state_data, tuple):
                return "(" + inner + ")"

        return str(state_data)

    def __init__(self, **kwargs: Any) -> None:
        """
        Initialize a complete automaton.
        """
        if not global_config.allow_mutable_automata:
            for attr_name, attr_value in kwargs.items():
                object.__setattr__(self, attr_name, freeze_value(attr_value))
        else:
            for attr_name, attr_value in kwargs.items():
                object.__setattr__(self, attr_name, attr_value)
        self.__post_init__()

    def __post_init__(self) -> None:
        """
        Perform post-initialization validation of the automaton.
        """
        if global_config.should_validate_automata:
            self.validate()

    @abc.abstractmethod
    def validate(self) -> None:
        """
        Raises an exception if this automaton is not internally consistent.

        Raises
        ------
        NotImplementedError
            If this method has not been implemented by a subclass.
        """
        raise NotImplementedError

    def __setattr__(self, name: str, value: Any) -> NoReturn:
        """
        Set custom setattr to make class immutable.

        Parameters
        ----------
        name : str
            The name of the attribute to set.
        value : Any
            The value to set the attribute to.

        Returns
        -------
        NoReturn
            This method does not return anything as it always raises.

        Raises
        ------
        AttributeError
            If this method is called.
        """
        raise AttributeError(f"This {type(self).__name__} is immutable")

    def __delattr__(self, name: str) -> None:
        """
        Set custom delattr to make class immutable.

        Parameters
        ----------
        name : str
            The name of the attribute to delete.

        Raises
        ------
        AttributeError
            If this method is called.
        """
        raise AttributeError(f"This {type(self).__name__} is immutable")

    def __getstate__(self) -> Any:
        """
        Return the object's state, described by its input parameters.

        Returns
        -------
        Any
            The object's state, described by its input parameters.
        """
        return self.input_parameters

    def __setstate__(self, d: Dict[str, Any]) -> None:
        """
        Restore the object state from its input parameters.

        Parameters
        ----------
        d : Dict[str, Any]
            The object's input parameters.
        """
        # Notice that the default __setstate__ method won't work
        #   because __setattr__ is disabled due to immutability
        self.__init__(**d)  # type: ignore

    @abc.abstractmethod
    def read_input_stepwise(self, input_str: str) -> Generator[Any, None, None]:
        """
        Return a generator that yields each step while reading input.

        Parameters
        ----------
        input_str : str
            The input string to read.

        Yields
        ------
        Generator[Any, None, None]
            A generator that yields the current configuration of the automaton
            after each step of reading input.

        Raises
        ------
        NotImplementedError
            If this method has not been implemented by a subclass.
        """
        raise NotImplementedError

    def read_input(self, input_str: str) -> AutomatonStateT:
        """
        Check if the given string is accepted by this automaton.

        Return the automaton's final configuration if this string is valid.

        Parameters
        ----------
        input_str : str
            The input string to check.

        Returns
        -------
        AutomatonStateT
            The final configuration of the automaton after reading the input.
        """
        # "Fast-forward" generator to get its final value
        for config in self.read_input_stepwise(input_str):
            pass
        return config

    def accepts_input(self, input_str: str) -> bool:
        """
        Return True if this automaton accepts the given input.

        Parameters
        ----------
        input_str : str
            The input string to check.

        Returns
        -------
        bool
            True if this automaton accepts the given input; False otherwise.
        """
        try:
            self.read_input(input_str)
            return True
        except exceptions.RejectionException:
            return False

    def _validate_initial_state(self) -> None:
        """
        Raise an error if the initial state is invalid.

        Raises
        ------
        exceptions.InvalidStateError
            If the initial state is invalid.
        """
        if self.initial_state not in self.states:
            raise exceptions.InvalidStateError(
                "{} is not a valid initial state".format(self.initial_state)
            )

    def _validate_initial_state_transitions(self) -> None:
        """
        Raise an error if the initial state has no transitions defined.

        Raises
        ------
        exceptions.MissingStateError
            If the initial state has no transitions defined.
        """
        if self.initial_state not in self.transitions and len(self.states) > 1:
            raise exceptions.MissingStateError(
                "initial state {} has no transitions defined".format(self.initial_state)
            )

    def _validate_final_states(self) -> None:
        """
        Raise an error if any final states are invalid.

        Raises
        ------
        exceptions.InvalidStateError
            If any final states are invalid.
        """
        invalid_states = self.final_states - self.states
        if invalid_states:
            raise exceptions.InvalidStateError(
                "final states are not valid ({})".format(
                    ", ".join(str(state) for state in invalid_states)
                )
            )

    @property
    def input_parameters(self) -> Dict[str, Any]:
        """
        Return the public attributes for this automaton.

        Returns
        -------
        Dict[str, Any]
            A dictionary mapping attribute names to their values.
        """
        return {
            attr_name: getattr(self, attr_name)
            for attr_name in self.__slots__
            if not attr_name.startswith("_")
        }

    def copy(self) -> Self:
        """
        Create a deep copy of the automaton.

        Returns
        -------
        Self
            A deep copy of the automaton.
        """
        return self.__class__(**self.input_parameters)

    # Format the given value for string output via repr() or str(); this exists
    # for the purpose of displaying

    def _get_repr_friendly_value(self, value: Any) -> Any:
        """
        A helper function to convert the given value / structure into a fully
        mutable one by recursively processing said structure and any of its
        members, unfreezing them along the way.

        Parameters
        ----------
        value : Any
            The value to convert.

        Returns
        -------
        Any
            A representation-friendly version of the given value.
        """
        if isinstance(value, frozenset):
            return set(value)
        elif isinstance(value, frozendict):
            return {
                dict_key: self._get_repr_friendly_value(dict_value)
                for dict_key, dict_value in value.items()
            }
        else:
            return value

    def __repr__(self) -> str:
        """
        Return a string representation of the automaton.

        Returns
        -------
        str
            A string representation of the automaton.
        """
        values = ", ".join(
            f"{attr_name}={self._get_repr_friendly_value(attr_value)!r}"
            for attr_name, attr_value in self.input_parameters.items()
        )
        return f"{self.__class__.__qualname__}({values})"

    def __contains__(self, item: Any) -> bool:
        """
        Returns whether the word is accepted by the automaton.

        Parameters
        ----------
        item : Any
            The item to check.

        Returns
        -------
        bool
            _description_
        """

        if not isinstance(item, str):
            return False

        return self.accepts_input(item)



"""Classes and functions for testing the Turing machine utility functions."""

import contextlib
import io
import unittest
from unittest.mock import MagicMock, call, patch

import automata.tm.tools as tmtools
from automata.tm.configuration import MTMConfiguration, TMConfiguration
from automata.tm.tape import TMTape


class TestTMTools(unittest.TestCase):
    """A test class for testing Turing machine utility functions."""

    config: TMConfiguration
    config2: MTMConfiguration

    def setUp(self) -> None:
        """Provide a configuration for testing."""
        self.config = TMConfiguration(
            "q2",
            TMTape(
                tape="abcdefghij",
                blank_symbol=".",
                current_position=2,
            ),
        )

        self.config2 = MTMConfiguration(
            "q1",
            (
                TMTape(
                    tape="abcdefghij",
                    blank_symbol=".",
                    current_position=2,
                ),
                TMTape(
                    tape="klmnopq",
                    blank_symbol=".",
                    current_position=5,
                ),
            ),
        )

    def test_repr_config(self) -> None:
        """Should return a string representation ot the given configuration."""
        self.assertEqual(
            repr(self.config), "TMConfiguration('q2', TMTape('abcdefghij', '.', 2))"
        )
        self.assertEqual(
            repr(self.config2),
            "MTMConfiguration('q1', (TMTape('abcdefghij', '.', 2), "
            + "TMTape('klmnopq', '.', 5)))",
        )

    def test_print_config(self) -> None:
        """Should print the given configuration to stdout."""
        out = io.StringIO()
        with contextlib.redirect_stdout(out):
            self.config.print()
        self.assertEqual(
            out.getvalue().rstrip(),
            "{}: {}\n{}".format("q2", "abcdefghij", "^".rjust(7)),
        )

    @patch("automata.tm.configuration.TMConfiguration.print")
    def test_print_configs(self, print_config: MagicMock) -> None:
        """Should print each machine configuration to stdout."""
        tape1 = TMTape(
            tape="01010101",
            blank_symbol=".",
            current_position=0,
        )
        tape2 = TMTape(
            tape="x1010101",
            blank_symbol=".",
            current_position=-1,
        )
        tape3 = TMTape(
            tape="yx1010101",
            blank_symbol=".",
            current_position=-2,
        )
        configs = [
            TMConfiguration("q0", tape1),
            TMConfiguration("q1", tape2),
            TMConfiguration("q2", tape3),
            MTMConfiguration("q1", (tape1, tape2, tape3)),
        ]
        out = io.StringIO()
        with contextlib.redirect_stdout(out):
            tmtools.print_configs(configs)  # type: ignore
            self.assertEqual(print_config.call_args_list, [call(), call(), call()])

    def test_tape_iteration(self) -> None:
        """Should be able to iterate over a Turing machine tape."""
        tape = TMTape(
            tape="abcdef",
            blank_symbol=".",
            current_position=2,
        )
        self.assertEqual(tuple(tape), ("a", "b", "c", "d", "e", "f"))

    def test_get_symbols_as_str(self) -> None:
        """Should print tape contents as a string without spaces."""
        tape = TMTape(
            tape="abcdef",
            blank_symbol=".",
            current_position=2,
        )
        self.assertEqual(tape.get_symbols_as_str(), "abcdef")



"""Classes and functions for testing the behavior of NTMs."""

import types
from unittest.mock import MagicMock, patch

from frozendict import frozendict

import automata.base.exceptions as exceptions
import automata.tm.exceptions as tm_exceptions
import tests.test_tm as test_tm
from automata.tm.ntm import NTM


class TestNTM(test_tm.TestTM):
    """A test class for testing nondeterministic Turing machines."""

    def test_init_ntm(self) -> None:
        """Should copy NTM if passed into NTM constructor."""
        new_dtm = self.ntm1.copy()
        self.assertIsNot(new_dtm, self.ntm1)

    def test_init_ntm_missing_formal_params(self) -> None:
        """Should raise an error if formal NTM parameters are missing."""
        with self.assertRaises(TypeError):
            NTM(  # type: ignore
                states={"q0", "q1", "q2", "q3", "q4"},
                input_symbols={"0", "1"},
                tape_symbols={"0", "1", "x", "y", "."},
                initial_state="q0",
                blank_symbol=".",
                final_states={"q4"},
            )

    @patch("automata.tm.ntm.NTM.validate")
    def test_init_validation(self, validate: MagicMock) -> None:
        """Should validate NTM when initialized."""
        self.ntm1.copy()
        validate.assert_called_once_with()

    def test_copy_ntm(self) -> None:
        """Should create exact copy of NTM if copy() method is called."""
        new_ntm = self.ntm1.copy()
        self.assertIsNot(new_ntm, self.ntm1)

    def test_ntm_immutable_attr_set(self) -> None:
        with self.assertRaises(AttributeError):
            self.ntm1.states = set()

    def test_ntm_immutable_attr_del(self) -> None:
        with self.assertRaises(AttributeError):
            del self.ntm1.states

    def test_ntm_immutable_dict(self) -> None:
        """Should create a DPDA whose contents are fully immutable/hashable"""
        self.assertIsInstance(hash(frozendict(self.ntm1.input_parameters)), int)

    def test_validate_input_symbol_subset(self) -> None:
        """Should raise error if input symbols are not a strict superset of tape
        symbols."""
        with self.assertRaises(exceptions.MissingSymbolError):
            NTM(
                states={"q0", "q1", "q2"},
                input_symbols={"0"},
                tape_symbols={"0"},
                transitions={
                    "q0": {"0": {("q1", "0", "R")}},
                    "q1": {"0": {("q2", "0", "L")}},
                },
                initial_state="q0",
                blank_symbol=".",
                final_states={"q2"},
            )

    def test_validate_invalid_transition_state(self) -> None:
        """Should raise error if a transition state is invalid."""
        with self.assertRaises(exceptions.InvalidStateError):
            NTM(
                states={"q0", "q1", "q2"},
                input_symbols={"0"},
                tape_symbols={"0", "."},
                transitions={
                    "q0": {"0": {("q1", "0", "R")}},
                    "q1": {"0": {("q2", "0", "L")}},
                    "q4": {"0": {("q1", "0", "R")}},
                },
                initial_state="q0",
                blank_symbol=".",
                final_states={"q2"},
            )

    def test_validate_invalid_transition_symbol(self) -> None:
        """Should raise error if a transition symbol is invalid."""
        with self.assertRaises(exceptions.InvalidSymbolError):
            NTM(
                states={"q0", "q1", "q2"},
                input_symbols={"0"},
                tape_symbols={"0", "."},
                transitions={
                    "q0": {"0": {("q1", "0", "R")}, "3": {("q0", "0", "R")}},
                    "q1": {"0": {("q2", "0", "L")}},
                },
                initial_state="q0",
                blank_symbol=".",
                final_states={"q2"},
            )

    def test_validate_invalid_transition_result_state(self) -> None:
        """Should raise error if a transition result state is invalid."""
        with self.assertRaises(exceptions.InvalidStateError):
            NTM(
                states={"q0", "q1", "q2"},
                input_symbols={"0"},
                tape_symbols={"0", "."},
                transitions={
                    "q0": {"0": {("q1", "0", "R")}, ".": {("q4", ".", "R")}},
                    "q1": {"0": {("q2", "0", "L")}},
                },
                initial_state="q0",
                blank_symbol=".",
                final_states={"q2"},
            )

    def test_validate_invalid_transition_result_symbol(self) -> None:
        """Should raise error if a transition result symbol is invalid."""
        with self.assertRaises(exceptions.InvalidSymbolError):
            NTM(
                states={"q0", "q1", "q2"},
                input_symbols={"0"},
                tape_symbols={"0", "."},
                transitions={
                    "q0": {"0": {("q1", "0", "R")}, ".": {("q2", "5", "R")}},
                    "q1": {"0": {("q2", "0", "L")}},
                },
                initial_state="q0",
                blank_symbol=".",
                final_states={"q2"},
            )

    def test_validate_invalid_transition_result_direction(self) -> None:
        """Should raise error if a transition result direction is invalid."""
        with self.assertRaises(tm_exceptions.InvalidDirectionError):
            NTM(
                states={"q0", "q1", "q2"},
                input_symbols={"0"},
                tape_symbols={"0", "."},
                transitions={
                    "q0": {"0": {("q1", "0", "U")}},  # type: ignore
                    "q1": {"0": {("q2", "0", "L")}},
                },
                initial_state="q0",
                blank_symbol=".",
                final_states={"q2"},
            )

    def test_validate_invalid_initial_state(self) -> None:
        """Should raise error if the initial state is invalid."""
        with self.assertRaises(exceptions.InvalidStateError):
            NTM(
                states={"q0", "q1", "q2"},
                input_symbols={"0"},
                tape_symbols={"0", "."},
                transitions={
                    "q0": {"0": {("q1", "0", "R")}},
                    "q1": {"0": {("q2", "0", "L")}},
                },
                initial_state="q5",
                blank_symbol=".",
                final_states={"q2"},
            )

    def test_validate_initial_state_transitions(self) -> None:
        """Should raise error if the initial state has no transitions."""
        with self.assertRaises(exceptions.MissingStateError):
            NTM(
                states={"q0", "q1", "q2"},
                input_symbols={"0"},
                tape_symbols={"0", "."},
                transitions={"q1": {"0": {("q2", "0", "L")}}},
                initial_state="q0",
                blank_symbol=".",
                final_states={"q2"},
            )

    def test_validate_nonfinal_initial_state(self) -> None:
        """Should raise error if the initial state is a final state."""
        with self.assertRaises(exceptions.InitialStateError):
            NTM(
                states={"q0", "q1", "q2"},
                input_symbols={"0"},
                tape_symbols={"0", "."},
                transitions={
                    "q0": {"0": {("q1", "0", "R")}},
                    "q1": {"0": {("q2", "0", "L")}},
                },
                initial_state="q0",
                blank_symbol=".",
                final_states={"q0"},
            )

    def test_validate_invalid_final_state(self) -> None:
        """Should raise error if the final state is invalid."""
        with self.assertRaises(exceptions.InvalidStateError):
            NTM(
                states={"q0", "q1", "q2"},
                input_symbols={"0"},
                tape_symbols={"0", "."},
                transitions={
                    "q0": {"0": {("q1", "0", "R")}},
                    "q1": {"0": {("q2", "0", "L")}},
                },
                initial_state="q0",
                blank_symbol=".",
                final_states={"q5"},
            )

    def test_validate_invalid_final_state_non_str(self) -> None:
        """Should raise InvalidStateError even for non-string final states."""
        with self.assertRaises(exceptions.InvalidStateError):
            NTM(
                states={"q0", "q1", "q2"},
                input_symbols={"0"},
                tape_symbols={"0", "."},
                transitions={
                    "q0": {"0": {("q1", "0", "R")}},
                    "q1": {"0": {("q2", "0", "L")}},
                },
                initial_state="q0",
                blank_symbol=".",
                final_states={5},
            )

    def test_validate_final_state_transitions(self) -> None:
        """Should raise error if a final state has any transitions."""
        with self.assertRaises(exceptions.FinalStateError):
            NTM(
                states={"q0", "q1", "q2"},
                input_symbols={"0"},
                tape_symbols={"0", "."},
                transitions={
                    "q0": {"0": {("q1", "0", "R")}},
                    "q1": {"0": {("q2", "0", "L")}},
                    "q2": {"0": {("q1", "0", "L")}},
                },
                initial_state="q0",
                blank_symbol=".",
                final_states={"q2"},
            )

    def test_read_input_accepted(self) -> None:
        """Should return correct state if acceptable TM input is given."""
        final_config = self.ntm1.read_input("00120001111").pop()
        self.assertEqual(final_config.state, "q3")
        self.assertEqual(str(final_config.tape), "TMTape('00120001111.', '.', 11)")

    def test_read_input_step(self) -> None:
        """Should return validation generator if step flag is supplied."""
        validation_generator = self.ntm1.read_input_stepwise("00120001111")
        self.assertIsInstance(validation_generator, types.GeneratorType)
        configs = list(validation_generator)
        first_config = configs[0].pop()
        self.assertEqual(first_config.state, "q0")
        self.assertEqual(str(first_config.tape), "TMTape('00120001111', '.', 0)")
        last_config = configs[-1].pop()
        self.assertEqual(last_config.state, "q3")
        self.assertEqual(str(last_config.tape), "TMTape('00120001111.', '.', 11)")

    def test_read_input_rejection(self) -> None:
        """Should raise error if the machine halts."""
        with self.assertRaises(exceptions.RejectionException):
            self.ntm1.read_input("0")

    def test_read_input_rejection_invalid_symbol(self) -> None:
        """Should raise error if an invalid symbol is read."""
        with self.assertRaises(exceptions.RejectionException):
            self.ntm1.read_input("02")

    def test_accepts_input_true(self) -> None:
        """Should return False if DTM input is not accepted."""
        self.assertTrue(self.dtm1.accepts_input("00001111"))

    def test_accepts_input_false(self) -> None:
        """Should return False if DTM input is rejected."""
        self.assertFalse(self.dtm1.accepts_input("000011"))



"""Global configuration for the library and its primitives"""

should_validate_automata: bool = True
# When set to True, it disables the freeze_value step
#   -> You must guarantee that your code does not modify the automata
allow_mutable_automata: bool = False
